<?php

namespace Drupal\openid_connect_akamai\Plugin\OpenIDConnectClient;

use Drupal\Component\Utility\Random;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\Routing\TrustedRedirectResponse;
use Drupal\Core\Url;
use Drupal\openid_connect\Plugin\OpenIDConnectClientBase;
use Drupal\openid_connect\StateToken;
use Exception;

/**
 * Akamai OpenID Connect client.
 *
 * Implements OpenID Connect Client plugin for Akamai(Janrain).
 *
 * @OpenIDConnectClient(
 *   id = "akamai",
 *   label = @Translation("Akamai(Janrain)")
 * )
 */
class Akamai extends OpenIDConnectClientBase {

  /**
   * Using the SHA256 hashing function.
   */
  const CODE_CHALLENGE_METHOD = 'sha256';
  const CODE_CHALLENGE_METHOD_AKAMAI_NAME = 'S256';

  /**
   * The name of the session variable in which the "code verifier" is stored.
   */
  const CODE_VERIFIER_VARIABLE = 'akamai_auth_code_verifier';

  /**
   * Default email login.
   *
   * @var null
   */
  private $emailForAuthorize = NULL;

  /**
   * {@inheritdoc}
   */
  public function authorize($scope = 'openid email') {
    // Use Akamai specific authorisations.
    $language_none = \Drupal::languageManager()
      ->getLanguage(LanguageInterface::LANGCODE_NOT_APPLICABLE);
    $redirect_uri = Url::fromRoute(
      'openid_connect.redirect_controller_redirect',
      [
        'client_name' => $this->pluginId,
      ],
      [
        'absolute' => TRUE,
        'language' => $language_none,
      ]
    )->toString(TRUE);

    $url_options = [
      'query' => [
        'client_id' => $this->configuration['client_id'],
        'response_type' => 'code',
        'scope' => $scope,
        'redirect_uri' => $redirect_uri->getGeneratedUrl(),
        'state' => StateToken::create(),
        'code_challenge' => $this->getCodeChallenge(),
        'code_challenge_method' => self::CODE_CHALLENGE_METHOD_AKAMAI_NAME,
      ],
    ];

    if ($this->emailForAuthorize) {
      $url_options['query']['oid_email'] = $this->emailForAuthorize;
      $this->emailForAuthorize = NULL;
    }

    $endpoints = $this->getEndpoints();
    // Clear _GET['destination'] because we need to override it.
    $this->requestStack->getCurrentRequest()->query->remove('destination');
    $authorization_endpoint = Url::fromUri($endpoints['authorization'], $url_options)->toString(TRUE);

    $response = new TrustedRedirectResponse($authorization_endpoint->getGeneratedUrl());
    // We can't cache the response, since this will prevent the state to be
    // added to the session. The kill switch will prevent the page getting
    // cached for anonymous users when page cache is active.
    \Drupal::service('page_cache_kill_switch')->trigger();

    return $response;
  }

  /**
   * {@inheritdoc}
   */
  public function authorizeByEmail($scopes, $email) {
    $this->emailForAuthorize = $email;
    return $this->authorize($scopes);
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['customer_id'] = [
      '#title' => $this->t('Customer ID'),
      '#type' => 'textfield',
      '#default_value' => $this->configuration['customer_id'],
    ];

    $form['client_secret']['#disabled'] = TRUE;
    $form['client_secret']['#type'] = 'hidden';

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function getEndpoints() {
    return [
      'authorization' => 'https://v1.api.us.janrain.com/' . $this->configuration['customer_id'] . '/login/authorize',
      'token' => 'https://v1.api.us.janrain.com/' . $this->configuration['customer_id'] . '/login/token',
      'userinfo' => 'https://v1.api.us.janrain.com/' . $this->configuration['customer_id'] . '/profiles/oidc/userinfo',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function retrieveTokens($authorization_code) {
    // Exchange `code` for access token and ID token.
    $language_none = \Drupal::languageManager()
      ->getLanguage(LanguageInterface::LANGCODE_NOT_APPLICABLE);
    $redirect_uri = Url::fromRoute(
      'openid_connect.redirect_controller_redirect',
      [
        'client_name' => $this->pluginId,
      ],
      [
        'absolute' => TRUE,
        'language' => $language_none,
      ]
    )->toString();
    $endpoints = $this->getEndpoints();

    $request_options = [
      'form_params' => [
        'code' => $authorization_code,
        'client_id' => $this->configuration['client_id'],
        'client_secret' => $this->configuration['client_secret'],
        'redirect_uri' => $redirect_uri,
        'grant_type' => 'authorization_code',
        'code_verifier' => $this->getCodeVerifier(),
      ],
      'headers' => [
        'Accept' => 'application/json',
      ],
    ];

    /* @var \GuzzleHttp\ClientInterface $client */
    $client = $this->httpClient;
    try {
      $response = $client->post($endpoints['token'], $request_options);
      $response_data = json_decode((string) $response->getBody(), TRUE);

      // Expected result.
      $tokens = [
        'id_token' => isset($response_data['id_token']) ? $response_data['id_token'] : NULL,
        'access_token' => isset($response_data['access_token']) ? $response_data['access_token'] : NULL,
      ];
      if (array_key_exists('expires_in', $response_data)) {
        $tokens['expire'] = REQUEST_TIME + $response_data['expires_in'];
      }
      if (array_key_exists('refresh_token', $response_data)) {
        $tokens['refresh_token'] = $response_data['refresh_token'];
      }
      return $tokens;
    }
    catch (Exception $e) {
      $variables = [
        '@message' => 'Could not retrieve tokens',
        '@error_message' => $e->getMessage(),
      ];
      $this->loggerFactory->get('openid_connect_' . $this->pluginId)
        ->error('@message. Details: @error_message', $variables);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function retrieveUserInfo($access_token) {
    $userinfo = parent::retrieveUserInfo($access_token);
    return $userinfo;
  }

  /**
   * Create a “code verifier” and Code Challenge(PKCE).
   *
   * Code verifier a random string of 43 to 128(126 for Drupal Generator)
   * characters.
   */
  protected function getCodeChallenge() {
    $random = new Random();
    $code_verifier = $random->name(mt_rand(43, 126));

    $session = \Drupal::request()->getSession();
    $session->set(self::CODE_VERIFIER_VARIABLE, $code_verifier);

    // Then base64-url encodes the hashed string
    // to create a “code challenge” string.
    $hash = strtoupper(hash(self::CODE_CHALLENGE_METHOD, $code_verifier));
    return str_replace(['+', '/', '='], ['-', '_', ''], base64_encode(pack('H*', $hash)));
  }

  /**
   * {@inheritdoc}
   */
  protected function getCodeVerifier() {
    $session = \Drupal::request()->getSession();
    return $session->get(self::CODE_VERIFIER_VARIABLE);
  }

}
